/* 
 * Copyright (C) 2004      Disch
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

//////////////////////////////////////////////////////////////////////////
//
//  Wave_FDS.h

static const int ModulationTable[8] = {0,1,2,4,0,-4,-2,-1};

struct FDSWave
{
	/*	Envelope Unit	*/
	BYTE		bEnvelopeEnable;
	BYTE		nEnvelopeSpeed;

	/*	Volume Envelope	*/
	BYTE		nVolEnv_Mode;
	BYTE		nVolEnv_Decay;
	BYTE		nVolEnv_Gain;
	int			nVolEnv_Timer;
	int			nVolEnv_Count;
	BYTE		nVolume;
	BYTE		bVolEnv_On;

	/*	Sweep Envenlope	*/
	BYTE		nSweep_Mode;
	BYTE		nSweep_Decay;
	int			nSweep_Timer;
	int			nSweep_Count;
	BYTE		nSweep_Gain;
	BYTE		bSweepEnv_On;

	/*	Effector / LFO / Modulation	Unit	*/
	int			nSweepBias;
	BYTE		bLFO_Enabled;
	union TWIN		nLFO_Freq;
	float		fLFO_Timer;
	float		fLFO_Count;
	BYTE		nLFO_Addr;
	BYTE		nLFO_Table[0x40];
	BYTE		bLFO_On;

	/*	Main Output		*/
	BYTE		nMainVolume;
	BYTE		bEnabled;
	union TWIN		nFreq;
	float		fFreqCount;
	BYTE		nMainAddr;
	BYTE		nWaveTable[0x40];
	BYTE		bWaveWrite;
	BYTE		bMain_On;

	/*	Output and Downsampling	*/
	short		nOutputTable_L[4][0x21][0x40];
	short		nOutputTable_R[4][0x21][0x40];
	int			nMixL;
	int			nMixR;

	/*	Inversion				*/
	BYTE		bInvert;

	/*	Pop Reducer				*/
	BYTE		bPopReducer;
	BYTE		nPopOutput;
	int			nPopCount;
	
};

extern struct FDSWave mWave_FDS;
#define this mWave_FDS
	
	///////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////
	//  Functions
	///////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////

	inline void Wave_FDS_DoTicks(int ticks,BYTE mix)
	{
		int mn;
		float freq;

		while(ticks)
		{
			mn = ticks;
			if(this.bVolEnv_On)						mn = min(mn,this.nVolEnv_Count);
			if(this.bSweepEnv_On)					mn = min(mn,this.nSweep_Count);
			if(this.bLFO_On)							mn = (int)min(mn,(this.fLFO_Count + 1));
			if(this.bMain_On)						mn = (int)min(mn,(this.fFreqCount + 1));
			else if(this.bPopReducer && this.nPopOutput)	mn = min(mn,this.nPopCount);
			ticks -= mn;

			/*	Volume Envelope Unit	*/
			if(this.bVolEnv_On)
			{
				this.nVolEnv_Count -= mn;
				if(this.nVolEnv_Count <= 0)
				{
					this.nVolEnv_Count += this.nVolEnv_Timer;
					if(this.nVolEnv_Mode){	if(this.nVolEnv_Gain < 0x20)	this.nVolEnv_Gain++; }
					else			{	if(this.nVolEnv_Gain)		this.nVolEnv_Gain--; }
				}
			}
			
			/*	Sweep Envelope Unit	*/
			if(this.bSweepEnv_On)
			{
				this.nSweep_Count -= mn;
				if(this.nSweep_Count <= 0)
				{
					this.nSweep_Count += this.nSweep_Timer;
					if(this.nSweep_Mode)	{	if(this.nSweep_Gain < 0x20)	this.nSweep_Gain++; }
					else			{	if(this.nSweep_Gain)			this.nSweep_Gain--; }
				}
			}

			/*	Effector / LFO		*/
			int		subfreq = 0;
			if(this.bLFO_On)
			{
				this.fLFO_Count -= mn;
				if(this.fLFO_Count <= 0)
				{
					this.fLFO_Count += this.fLFO_Timer;
					if(this.nLFO_Table[this.nLFO_Addr] == 4)	this.nSweepBias = 0;
					else							this.nSweepBias += ModulationTable[ this.nLFO_Table[this.nLFO_Addr] ];
					this.nLFO_Addr = (this.nLFO_Addr + 1) & 0x3F;
				}

				while(this.nSweepBias >  63)	this.nSweepBias -= 128;
				while(this.nSweepBias < -64)	this.nSweepBias += 128;

				register int temp = this.nSweepBias * this.nSweep_Gain;
				if(temp & 0x0F)
				{
					temp /= 16;
					if(this.nSweepBias < 0)	temp--;
					else				temp += 2;
				}
				else
					temp /= 16;

				if(temp > 193)	temp -= 258;
				if(temp < -64)	temp += 256;

				subfreq = this.nFreq.W * temp / 64;
			}

			/*	Main Unit		*/
			if(this.bMain_On)
			{
				if(mix)
				{
					this.nMixL += this.nOutputTable_L[this.nMainVolume][this.nVolume][ this.nWaveTable[this.nMainAddr] ] * mn;
					this.nMixR += this.nOutputTable_R[this.nMainVolume][this.nVolume][ this.nWaveTable[this.nMainAddr] ] * (this.bInvert ? -mn : mn);
				}

				if((subfreq + this.nFreq.W) > 0)
				{
					freq = 65536.0f / (subfreq + this.nFreq.W);

					this.fFreqCount -= mn;
					if(this.fFreqCount <= 0)
					{
						this.fFreqCount += freq;

						this.nMainAddr = (this.nMainAddr + 1) & 0x3F;
						this.nPopOutput = this.nWaveTable[this.nMainAddr];
						if(!this.nMainAddr)
						{
							if(this.nVolEnv_Gain < 0x20)		this.nVolume = this.nVolEnv_Gain;
							else						this.nVolume = 0x20;
						}
					}
				}
				else
					this.fFreqCount = this.fLFO_Count;
			}
			else if(this.bPopReducer && this.nPopOutput)
			{
				if(mix)
				{
					this.nMixL += this.nOutputTable_L[this.nMainVolume][this.nVolume][ this.nPopOutput ] * mn;
					this.nMixR += this.nOutputTable_R[this.nMainVolume][this.nVolume][ this.nPopOutput ] * (this.bInvert ? -mn : mn);
				}

				this.nPopCount -= mn;
				if(this.nPopCount <= 0)
				{
					this.nPopCount += 500;
					this.nPopOutput--;
					if(!this.nPopOutput)
						this.nMainAddr = 0;
				}
			}
		}
	}
	
	//////////////////////////////////////////////////////////////////////////
	//  Mixing

	inline void Wave_FDS_Mix_Mono(int * mix,int downsample)
	{
		*mix += (this.nMixL / downsample);
		this.nMixL = 0;
	}
	
	inline void Wave_FDS_Mix_Stereo(int * mixL,int * mixR,int downsample)
	{
		*mixL += (this.nMixL / downsample);
		*mixR += (this.nMixR / downsample);

		this.nMixL = this.nMixR = 0;
	}
#undef this
