/* 
 * Copyright (C) 2004      Disch
 * Copyright (C) 2006      hcs
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

//////////////////////////////////////////////////////////////////////////
//
//  Wave_TND.h
//
//    Because of inter-dependencies between channel output, the Triangle,
//  Noise, and DMC are all combined here in this class.

struct Wave_TND
{

	/*	Triangle			*/

	///////////////////////////////////
	// Programmable Timer
	union TWIN		nTriFreqTimer;
	int			nTriFreqCount;

	///////////////////////////////////
	// Length Counter
	BYTE		nTriLengthCount;
	BYTE		bTriLengthEnabled;
	BYTE		bTriChannelEnabled;

	///////////////////////////////////
	// Linear Counter
	BYTE		nTriLinearCount;
	BYTE		nTriLinearLoad;
	BYTE		bTriLinearHalt;
	BYTE		bTriLinearControl;

	///////////////////////////////////
	// Tri-Step Generator / Output
	BYTE		nTriStep;
	BYTE		nTriOutput;
	BYTE		bTriChannelMix;


	/*	Noise				*/

	///////////////////////////////////
	// Programmable Timer
	WORD		nNoiseFreqTimer;
	int			nNoiseFreqCount;

	///////////////////////////////////
	// Length Counter
	BYTE		nNoiseLengthCount;
	BYTE		bNoiseLengthEnabled;
	BYTE		bNoiseChannelEnabled;

	///////////////////////////////////
	// Volume / Decay
	BYTE		nNoiseVolume;
	BYTE		nNoiseDecayVolume;
	BYTE		bNoiseDecayEnable;
	BYTE		bNoiseDecayLoop;
	BYTE		nNoiseDecayTimer;
	BYTE		nNoiseDecayCount;

	///////////////////////////////////
	// Random Number Generator
	WORD		nNoiseRandomShift;
	BYTE		bNoiseRandomMode;			//1 = 32k, 6 = 93-bit
	BYTE		bNoiseRandomOut;
	BYTE		bNoiseChannelMix;


	/*	DMC					*/

	///////////////////////////////////
	// Play Mode
	BYTE		bDMCLoop;
	BYTE		bDMCIRQEnabled;
	BYTE		bDMCIRQPending;

	///////////////////////////////////
	// Address / DMA
	BYTE		nDMCDMABank_Load;
	WORD		nDMCDMAAddr_Load;
	BYTE		nDMCDMABank;
	WORD		nDMCDMAAddr;
	BYTE*		pDMCDMAPtr[8];

	//////////////////////////////////
	// Length / Input
	WORD		nDMCLength;
	WORD		nDMCBytesRemaining;
	BYTE		nDMCDelta;
	BYTE		nDMCDeltaBit;
	BYTE		bDMCDeltaSilent;
	BYTE		nDMCSampleBuffer;
	BYTE		bDMCSampleBufferEmpty;

	//////////////////////////////////
	// Frequency
	WORD		nDMCFreqTimer;
	WORD		nDMCFreqCount;

	//////////////////////////////////
	// Output
	BYTE		bDMCActive;
	BYTE		nDMCOutput;
	BYTE		bDMCChannelMix;

	/*	All's Output table			*/
	short*		nOutputTable_L;
	short*		nOutputTable_R;
	int			nMixL;
	int			nMixR;

	///////////////////////////////////
	// Inverting
	BYTE		bInvert;
	BYTE		bDoInvert;
	WORD		nInvertFreqCutoff_Noise;
	WORD		nInvertFreqCutoff_Tri;
	
	
};
extern struct Wave_TND mWave_TND;

#define this mWave_TND

///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////
//  Functions
///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////

inline void Wave_TND_ClockMajor()		// decay (noise), linear (tri)
{
	//noise's decay
	if(this.nNoiseDecayCount)
		this.nNoiseDecayCount--;
	else
	{
		this.nNoiseDecayCount = this.nNoiseDecayTimer;
		if(this.nNoiseDecayVolume)
			this.nNoiseDecayVolume--;
		else
		{
			if(this.bNoiseDecayLoop)
				this.nNoiseDecayVolume = 0x0F;
		}

		if(this.bNoiseDecayEnable)
    		this.nNoiseVolume = this.nNoiseDecayVolume;
	}

	//triangle's linear
	if(this.bTriLinearHalt)
		this.nTriLinearCount = this.nTriLinearLoad;
	else if(this.nTriLinearCount)
		this.nTriLinearCount--;

	if(!this.bTriLinearControl)
		this.bTriLinearHalt = 0;
}

inline void Wave_TND_ClockMinor()		//length
{
	if(this.bNoiseLengthEnabled && this.nNoiseLengthCount)
		this.nNoiseLengthCount--;
		
	if(this.bTriLengthEnabled && this.nTriLengthCount)
		this.nTriLengthCount--;
}

inline int Wave_TND_DoTicks(int ticks)		//returns number of burned cycles (burned by DMC's DMA)
{
	register int mn;
	register int out;
	int burnedcycles = 0;

	while(ticks)
	{
		mn = this.nNoiseFreqCount;
		if(this.nTriFreqTimer.W > 8)
			mn = min(mn,this.nTriFreqCount);
		if(this.bDMCActive)
			mn = min(mn,this.nDMCFreqCount);
		mn = min(mn,ticks);
		ticks -= mn;

		this.nNoiseFreqCount -= mn;
			
		if(this.nTriFreqTimer.W > 8)
			this.nTriFreqCount -= mn;

		if(!this.bTriChannelMix)												out = 0;
		else															out = this.nTriOutput << 11;

		if(this.bNoiseRandomOut && this.nNoiseLengthCount && this.bNoiseChannelMix)	out |= this.nNoiseVolume << 7;

		if(this.bDMCChannelMix)												out |= this.nDMCOutput;


		this.nMixL += this.nOutputTable_L[out] * mn;
		if(this.bDoInvert & 4)					this.nMixR -= this.nOutputTable_R[((this.bDoInvert ^ 7) << 15) | out] * mn;
		else								this.nMixR += this.nOutputTable_R[( this.bDoInvert      << 15) | out] * mn;

		/*	Tri				*/
		if(!this.nTriFreqCount)
		{
			this.nTriFreqCount = this.nTriFreqTimer.W + 1;
			if(this.nTriLengthCount && this.nTriLinearCount)
			{
				this.nTriStep = (this.nTriStep + 1) & 0x1F;

				if(this.nTriStep & 0x10)		this.nTriOutput = this.nTriStep ^ 0x1F;
				else					this.nTriOutput = this.nTriStep;

				if(!this.nTriStep)
				{
					if(this.bInvert & 1)
					{
						if(this.nTriFreqTimer.W <= this.nInvertFreqCutoff_Tri)	this.bDoInvert |= 1;
						else											this.bDoInvert &= 6;
					}
					else	this.bDoInvert &= 6;
				}
			}
		}

		/*	Noise			*/
		if(!this.nNoiseFreqCount)
		{
			this.nNoiseFreqCount = this.nNoiseFreqTimer;
			this.nNoiseRandomShift <<= 1;
			this.bNoiseRandomOut = (((this.nNoiseRandomShift << this.bNoiseRandomMode) ^ this.nNoiseRandomShift) & 0x8000 ) ? 1 : 0;
			if(this.bNoiseRandomOut)
				this.nNoiseRandomShift |= 0x01;
			else
			{
				if(this.bInvert & 2)
				{
					if(this.nNoiseFreqTimer <= this.nInvertFreqCutoff_Noise)	this.bDoInvert |= 2;
					else											this.bDoInvert &= 5;
				}
				else	this.bDoInvert &= 5;
			}
		}


		/*	DMC				*/
		if(this.bInvert & 4)		this.bDoInvert |= 4;
		else				this.bDoInvert &= 3;
		if(this.bDMCActive)
		{
			this.nDMCFreqCount -= mn;
			if(this.nDMCFreqCount > 0)
				continue;

			this.nDMCFreqCount = this.nDMCFreqTimer;

			if(this.bDMCSampleBufferEmpty && this.nDMCBytesRemaining)
			{
				burnedcycles += 4;		//4 cycle burn!
				this.nDMCSampleBuffer = this.pDMCDMAPtr[this.nDMCDMABank][this.nDMCDMAAddr];
				this.nDMCDMAAddr++;
				if(this.nDMCDMAAddr & 0x1000)
				{
					this.nDMCDMAAddr &= 0x0FFF;
					this.nDMCDMABank = (this.nDMCDMABank + 1) & 0x07;
				}

				this.bDMCSampleBufferEmpty = 0;
				this.nDMCBytesRemaining--;
				if(!this.nDMCBytesRemaining)
				{
					if(this.bDMCLoop)
					{
						this.nDMCDMABank = this.nDMCDMABank_Load;
						this.nDMCDMAAddr = this.nDMCDMAAddr_Load;
						this.nDMCBytesRemaining = this.nDMCLength;
					}
					else if(this.bDMCIRQEnabled)
					this.bDMCIRQPending = 1;
				}
			}

			if(!this.nDMCDeltaBit)
			{
				this.nDMCDeltaBit = 8;
				this.bDMCDeltaSilent = this.bDMCSampleBufferEmpty;
				this.nDMCDelta = this.nDMCSampleBuffer;
				this.bDMCSampleBufferEmpty = 1;
			}
				
			if(this.nDMCDeltaBit)
			{
				this.nDMCDeltaBit--;
				if(!this.bDMCDeltaSilent)
				{
					if(this.nDMCDelta & 0x01)
					{
						if(this.nDMCOutput < 0x7E) this.nDMCOutput += 2;
					}
					else if(this.nDMCOutput > 1)	this.nDMCOutput -= 2;
				}
				this.nDMCDelta >>= 1;
			}

			if(!this.nDMCBytesRemaining && this.bDMCSampleBufferEmpty && this.bDMCDeltaSilent)
				this.bDMCActive = this.nDMCDeltaBit = 0;
		}
	}
	return burnedcycles;
}
	
inline void Wave_TND_Mix_Mono(int * mix,int downsample)
{
	*mix += (this.nMixL / downsample);
	this.nMixL = 0;
}
	
inline void Wave_TND_Mix_Stereo(int * mixL,int * mixR,int downsample)
{
	*mixL += (this.nMixL / downsample);
	*mixR += (this.nMixR / downsample);

	this.nMixL = this.nMixR = 0;
}

#undef this
